#pragma once

#include "drake/common/drake_assert.h"

// This file provides backwards compatiblity with an old include path. In new
// code, please include drake_assert.h instead. TODO(jwnimmer-tri) On or after
// 2027-01-01, schedule this file for removal (with deprecation).
