# noqa: shebang
"""
Provides information on the external fork of `pybind11` used by `pydrake`.
"""

repository = "pybind/pybind11"
commit = "v2.13.6"
sha256 = "e08cb87f4773da97fa7b5f035de8763abc656d87d5773e62f6da0587d1f0ec20"
