#pragma once

// This header is generated from version_internal.h.in; do not edit it by hand.
//
// It defines the concrete Drake version values (from the same version stamp
// that populates drake-config.cmake) that drake/version.h consumes. These
// macros are for Drake's internal use only; downstream code should instead use
// the macros in drake/version.h.

#define DRAKE_INTERNAL_VERSION_STRING "0.0.20260818"
#define DRAKE_INTERNAL_VERSION_MAJOR 0
#define DRAKE_INTERNAL_VERSION_MINOR 0
#define DRAKE_INTERNAL_VERSION_PATCH 20260818
